//   GreKo chess engine
//   (c) 2002-2021 Vladimir Medvedev <vrm@bk.ru>
//   http://greko.su

#include "eval.h"
#include "eval_params.h"
#include "utils.h"

extern string g_weightsFile;

const int PAWN_HASHTABLE_BITS = 16;
const U64 PAWN_HASHTABLE_MASK = (U64(1) << PAWN_HASHTABLE_BITS) - 1;
PawnStruct g_pawnHash[PAWN_HASHTABLE_MASK + 1];

EVAL PSQ[14][64];

void InitPSQ(const vector<double>& x);
int KingPawnShield(const Position& pos, COLOR side, const PawnStruct& ps);
int KingPawnStorm(const Position& pos, COLOR side, const PawnStruct& ps);

void AddPsq(
	vector<double>& features,
	size_t index,
	COLOR side,
	FLD f)
{
	double UNIT = (1 - 2 * side);
	FLD f1 = FLIP_SIDE[side][f];

	switch (PSQ_LENGTH)
	{
	case 5:
		{
			double X = (Col(f1) - 3.5) / 3.5;
			double Y = (3.5 - Row(f1)) / 3.5;
			double X2 = X * X;
			double Y2 = Y * Y;
			double XY = X * Y;

			features[index] += UNIT;
			features[index + 1] += UNIT * X2;
			features[index + 2] += UNIT * X;
			features[index + 3] += UNIT * Y2;
			features[index + 4] += UNIT * Y;
			features[index + 5] += UNIT * XY;
		}
		break;
	case 12:
		{
			int col = (Col(f1) < 4)? Col(f1) : 7 - Col(f1);
			int row = Row(f1);
			features[index] += UNIT;
			features[index + 1 + col] += UNIT;
			features[index + 5 + row] += UNIT;
		}
		break;
	case 16:
		{
			int col = Col(f1);
			int row = Row(f1);
			features[index] += UNIT;
			features[index + 1 + col] += UNIT;
			features[index + 9 + row] += UNIT;
		}
		break;
	case 64:
		{
			features[index] += UNIT;
			features[index + 1 + f1] += UNIT;
		}
		break;
	}
}
////////////////////////////////////////////////////////////////////////////////

void AddQuadratic(
	vector<double>& features,
	size_t index,
	COLOR side,
	double z)
{
	double UNIT = (1 - 2 * side);

	features[index] += UNIT * z * z;
	features[index + 1] += UNIT * z;
}
////////////////////////////////////////////////////////////////////////////////

void AddLinear(
	vector<double>& features,
	size_t index,
	COLOR side,
	double z)
{
	double UNIT = (1 - 2 * side);
	features[index] += UNIT * z;
}
////////////////////////////////////////////////////////////////////////////////

int Distance(FLD f1, FLD f2)
{
	static const int dist[100] =
	{
		0, 1, 1, 1, 2, 2, 2, 2, 2, 3,
		3, 3, 3, 3, 3, 3, 4, 4, 4, 4,
		4, 4, 4, 4, 4, 5, 5, 5, 5, 5,
		5, 5, 5, 5, 5, 5, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 9, 9, 9, 9, 9, 9, 9, 9, 9,
		9, 9, 9, 9, 9, 9, 9, 9, 9, 9
	};

	int drow = Row(f1) - Row(f2);
	int dcol = Col(f1) - Col(f2);
	return dist[drow * drow + dcol * dcol];
}
////////////////////////////////////////////////////////////////////////////////

EVAL Evaluate(const Position& pos, EVAL alpha, EVAL beta)
{
	COLOR side = pos.Side();
	COLOR opp = side ^ 1;

	EVAL e = (EVAL)(pos.Score(side) - pos.Score(opp));

	if (e > 0 && pos.Bits(PAWN | side) == 0 && pos.MatIndex(side) < 5)
		e = 0; 
	if (e < 0 && pos.Bits(PAWN | opp) == 0 && pos.MatIndex(opp) < 5)
		e = 0;

	if (pos.Fifty() > 100)
		e = 0;
	else
		e = e * (100 - pos.Fifty()) / 100;

	return e;
}
////////////////////////////////////////////////////////////////////////////////

EVAL FastEval(const Position& pos)
{
	return (EVAL)(pos.Score(pos.Side()) - pos.Score(pos.Side() ^ 1));
}
////////////////////////////////////////////////////////////////////////////////

void GetFeaturesSide(const Position& pos, vector<double>& features, COLOR side, const PawnStruct& ps)
{
	U64 x;
	FLD f;

#define ADD_PSQ(tag)                                       \
{                                                          \
    AddPsq(features, tag, side, f);                        \
}

#define ADD_FEATURE(tag)                                   \
{                                                          \
	AddLinear(features, ##tag, side, 1);                   \
}

#define ADD_FEATURE_VECTOR(tag, index)                     \
{                                                          \
	AddLinear(features, ##tag + index, side, 1);           \
}

#ifdef FEATURES_LINEAR
    #define ADD_FEATURE_SCALED(tag, value, maxValue)       \
    {                                                      \
        double z = double(value) / maxValue;               \
        AddLinear(features, ##tag, side, z);               \
    }
#endif

#ifdef FEATURES_QUADRATIC
    #define ADD_FEATURE_SCALED(tag, value, maxValue)       \
    {                                                      \
        double z = double(value) / maxValue;               \
        AddQuadratic(features, ##tag, side, z);            \
    }
#endif

#ifdef FEATURES_TABULAR
    #define ADD_FEATURE_SCALED(tag, value, maxValue)       \
    {                                                      \
        AddLinear(features, ##tag + value, side, 1);       \
    }
#endif

	//
	//   PAWNS
	//

	U64 pawns = pos.Bits(PAWN | side);

	x = pawns;
	while (x)
	{
		f = PopLSB(x);
		ADD_PSQ(Pawn)
	}

	//
	//   KNIGHTS
	//

	x = pos.Bits(KNIGHT | side);
	while (x)
	{
		f = PopLSB(x);
		ADD_PSQ(Knight)
	}

	//
	//   BISHOPS
	//

	x = pos.Bits(BISHOP | side);
	while (x)
	{
		f = PopLSB(x);
		ADD_PSQ(Bishop)
	}

	//
	//   ROOKS
	//

	x = pos.Bits(ROOK | side);
	while (x)
	{
		f = PopLSB(x);
		ADD_PSQ(Rook)
	}

	//
	//   QUEENS
	//

	x = pos.Bits(QUEEN | side);
	while (x)
	{
		f = PopLSB(x);
		ADD_PSQ(Queen)
	}

	//
	//   KINGS
	//

	{
		f = pos.King(side);
		ADD_PSQ(King)
	}

#undef ADD_PSQ
#undef ADD_FEATURE
#undef ADD_FEATURE_VECTOR
#undef ADD_FEATURE_SCALED
}
////////////////////////////////////////////////////////////////////////////////

void GetFeatures(const Position& pos, vector<double>& features)
{
	features.clear();
	features.resize(NUMBER_OF_FEATURES);

	size_t index = pos.PawnHash() & PAWN_HASHTABLE_MASK;
	PawnStruct& ps = g_pawnHash[index];
	if (ps.pawnHash != pos.PawnHash())
		ps.Read(pos);

	GetFeaturesSide(pos, features, WHITE, ps);
	GetFeaturesSide(pos, features, BLACK, ps);
}
////////////////////////////////////////////////////////////////////////////////

void InitEval(const vector<double>& x)
{
	g_w = x;
	InitPSQ(x);

#define INIT_WEIGHT(var, tag) \
    var = x[tag];

#define INIT_VECTOR(table, tag)             \
for (size_t i = 0; i <= MAX_##table; ++i)   \
{                                           \
    table[i] = x[tag + i];                  \
}

#ifdef FEATURES_LINEAR
    #define INIT_WEIGHTS(table, tag)            \
    for (size_t i = 0; i <= MAX_##table; ++i)   \
    {                                           \
        double z = double(i) / MAX_##table;     \
        table[i] =                              \
            x[tag] * z;                         \
    }
#endif

#ifdef FEATURES_QUADRATIC
    #define INIT_WEIGHTS(table, tag)            \
    for (size_t i = 0; i <= MAX_##table; ++i)   \
    {                                           \
        double z = double(i) / MAX_##table;     \
        table[i] =                              \
            x[tag] * z * z +                    \
            x[tag + 1] * z;                     \
    }
#endif

#ifdef FEATURES_TABULAR
    #define INIT_WEIGHTS(table, tag)            \
        INIT_VECTOR(table, tag)
#endif

#undef INIT_WEIGHTS
}
////////////////////////////////////////////////////////////////////////////////

void InitEval()
{
	InitFeatures();

	vector<double> x;
	if (!ReadWeights(x, g_weightsFile))
	{
		SetDefaultWeights(x);
		WriteWeights(x, g_weightsFile);
	}
	InitEval(x);
}
////////////////////////////////////////////////////////////////////////////////

EVAL FieldValue(
	const vector<double>& x,
	size_t tag,
	FLD f)
{
	switch (PSQ_LENGTH)
	{
	case 5:
		{
			double X = (Col(f) - 3.5) / 3.5;
			double Y = (3.5 - Row(f)) / 3.5;
			double X2 = X * X;
			double Y2 = Y * Y;
			double XY = X * Y;

			return (EVAL)
				(x[tag] +
				x[tag + 1] * X2 +
				x[tag + 2] * X +
				x[tag + 3] * Y2 +
				x[tag + 4] * Y +
				x[tag + 5] * XY);
		}
		break;
	case 12:
		{
			int col = (Col(f) < 4)? Col(f) : 7 - Col(f);
			int row = Row(f);

			return (EVAL)
				(x[tag] +
				x[tag + 1 + col] +
				x[tag + 5 + row]);
		}
		break;
	case 16:
		{
			int col = Col(f);
			int row = Row(f);

			return (EVAL)
				(x[tag] +
				x[tag + 1 + col] +
				x[tag + 9 + row]);
		}
		break;
	case 64:
		{
			return (EVAL)
				(x[tag] +
				x[tag + 1 + f]);
		}
		break;
	default:
		return 0;
	}
}
////////////////////////////////////////////////////////////////////////////////

void InitPSQ(const vector<double>& x)
{
	for (FLD f = 0; f < 64; ++f)
	{
		PSQ[PW][f] = FieldValue(x, Pawn, f);
		PSQ[NW][f] = FieldValue(x, Knight, f);
		PSQ[BW][f] = FieldValue(x, Bishop, f);
		PSQ[RW][f] = FieldValue(x, Rook, f);
		PSQ[QW][f] = FieldValue(x, Queen, f);
		PSQ[KW][f] = FieldValue(x, King, f);

		PSQ[PB][FLIP[f]] = PSQ[PW][f];
		PSQ[NB][FLIP[f]] = PSQ[NW][f];
		PSQ[BB][FLIP[f]] = PSQ[BW][f];
		PSQ[RB][FLIP[f]] = PSQ[RW][f];
		PSQ[QB][FLIP[f]] = PSQ[QW][f];
		PSQ[KB][FLIP[f]] = PSQ[KW][f];
	}
}
////////////////////////////////////////////////////////////////////////////////

int KingPawnShield(const Position& pos, COLOR side, const PawnStruct& ps)
{
	FLD K = pos.King(side);
	int file = Col(K) + 1;
	
	const int bonus[2][8] =
	{
		{ 0, 0, 0, 0, 1, 2, 3, 0 },
		{ 0, 3, 2, 1, 0, 0, 0, 0 }
	};

	int shield = 0;
	for (int i = file - 1; i < file + 2; ++i)
	{
		int rank = ps.ranks[side][i];
		shield += bonus[side][rank];
	}
	return shield;
}
////////////////////////////////////////////////////////////////////////////////

int KingPawnStorm(const Position& pos, COLOR side, const PawnStruct& ps)
{
	FLD K = pos.King(side);
	int file = Col(K) + 1;
	
	const int penalty[2][8] =
	{
		{ 0, 0, 0, 0, 1, 2, 3, 3 },
		{ 3, 3, 2, 1, 0, 0, 0, 0 }
	};

	int storm = 0;
	for (int i = file - 1; i < file + 2; ++i)
	{
		int rank = ps.ranks[side ^ 1][i];
		storm += penalty[side][rank];
	}
	return storm;
}
////////////////////////////////////////////////////////////////////////////////

void PawnStruct::Clear()
{
	pawnHash = 0;
	passed = 0;
	doubled = 0;
	isolated = 0;
	backwards = 0;

	attBy[WHITE] = attBy[BLACK] = 0;
	safe[WHITE] = safe[BLACK] = LL(0xffffffffffffffff);

	memset(ranks[WHITE], 0, 10);
	memset(ranks[BLACK], 7, 10);
}
////////////////////////////////////////////////////////////////////////////////

void PawnStruct::Read(const Position& pos)
{
	// Algorithm from TSCP 1.81 by Tom Kerrigan

	Clear();
	pawnHash = pos.PawnHash();

	U64 x, y;
	FLD f;

	// first pass

	x = pos.Bits(PW);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);
		if (rank > ranks[WHITE][file])
			ranks[WHITE][file] = rank;

		attBy[WHITE] |= BB_PAWN_ATTACKS[f][WHITE];
		y = Left(BB_DIR[f][DIR_U]) | Right(BB_DIR[f][DIR_U]);
		safe[BLACK] &= ~y;
	}

	x = pos.Bits(PB);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);
		if (rank < ranks[BLACK][file])
			ranks[BLACK][file] = rank;

		attBy[BLACK] |= BB_PAWN_ATTACKS[f][BLACK];
		y = Left(BB_DIR[f][DIR_D]) | Right(BB_DIR[f][DIR_D]);
		safe[WHITE] &= ~y;
	}

	// second pass

	x = pos.Bits(PW);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);

		if (BB_DIR[f][DIR_U] & pos.Bits(PW))
			doubled |= BB_SINGLE[f];

		if (ranks[WHITE][file - 1] == 0 && ranks[WHITE][file + 1] == 0)
			isolated |= BB_SINGLE[f];
		else if (rank > ranks[WHITE][file - 1] && rank > ranks[WHITE][file + 1])
			backwards |= BB_SINGLE[f];

		if (rank < ranks[BLACK][file] && rank <= ranks[BLACK][file - 1] && rank <= ranks[BLACK][file + 1])
			passed |= BB_SINGLE[f];
	}

	x = pos.Bits(PB);
	while (x)
	{
		f = PopLSB(x);
		int file = Col(f) + 1;
		int rank = Row(f);

		if (BB_DIR[f][DIR_D] & pos.Bits(PB))
			doubled |= BB_SINGLE[f];

		if (ranks[BLACK][file - 1] == 7 && ranks[BLACK][file + 1] == 7)
			isolated |= BB_SINGLE[f];
		else if (rank < ranks[BLACK][file - 1] && rank < ranks[BLACK][file + 1])
			backwards |= BB_SINGLE[f];

		if (rank > ranks[WHITE][file] && rank >= ranks[WHITE][file - 1] && rank >= ranks[WHITE][file + 1])
			passed |= BB_SINGLE[f];
	}
}
////////////////////////////////////////////////////////////////////////////////
