//   GreKo chess engine
//   (c) 2002-2021 Vladimir Medvedev <vrm@bk.ru>
//   http://greko.su

#include "eval_params.h"
#include "notation.h"
#include "utils.h"

struct FeatureInfo
{
	FeatureInfo(const string& name_, size_t index_, size_t len_) :
		name(name_),
		index(index_),
		len(len_)
	{}

	string name;
	size_t index;
	size_t len;
};
vector<FeatureInfo> g_features;

vector<double> g_w;

size_t Pawn;
size_t Knight;
size_t Bishop;
size_t Rook;
size_t Queen;
size_t King;

size_t NUMBER_OF_FEATURES;
size_t maxTagLength = 0;

void InitFeatures()
{
	size_t index = 0;
	g_features.clear();

#define REGISTER_FEATURE(var, len)                            \
	var = index;                                              \
	g_features.push_back(FeatureInfo(#var, index, len));      \
	index += len;                                             \
	if (strlen(#var) > maxTagLength)                          \
	    maxTagLength = strlen(#var);

	REGISTER_FEATURE(Pawn, PSQ_LENGTH + 1)
	REGISTER_FEATURE(Knight, PSQ_LENGTH + 1)
	REGISTER_FEATURE(Bishop, PSQ_LENGTH + 1)
	REGISTER_FEATURE(Rook, PSQ_LENGTH + 1)
	REGISTER_FEATURE(Queen, PSQ_LENGTH + 1)
	REGISTER_FEATURE(King, PSQ_LENGTH + 1)

	NUMBER_OF_FEATURES = index;

#undef REGISTER_FEATURE
}
////////////////////////////////////////////////////////////////////////////////

string FeatureName(size_t index)
{
	for (size_t i = 0; i < g_features.size(); ++i)
	{
		if (g_features[i].index <= index &&
			index < g_features[i].index + g_features[i].len)
		{
			stringstream ss;
			ss << g_features[i].name <<
				"[" << index - g_features[i].index << "]";
			return ss.str();
		}
	}
	return "<unknown>";
}
////////////////////////////////////////////////////////////////////////////////

void SetDefaultWeights(vector<double>& x)
{
	x.clear();
	x.resize(NUMBER_OF_FEATURES);

	static const double src5[] =
	{
		98, 7, 7, 71, 73, -7, 352, -29, 21, -10, 60, 17, 370, -9, 3, -26, 24,
		2, 538, -28, -2, 8, 58, 3, 1025, -19, 26, 2, 60, 44, 0, 8, -10, 10,
		57, -8
	};

	static const double src12[] =
	{
		85, 14, 26, 13, 13, 0, 97, 46, 5, -12, -20, -22, 0, 229, 26, 54, 63,
		68, 13, 52, 75, 57, 38, 12, -8, -22, 255, 50, 61, 63, 64, 30, 37, 64,
		41, 36, 27, 11, -7, 381, 76, 94, 95, 102, 83, 91, 93, 69, 29, 6, -5,
		0, 730, 167, 175, 183, 185, 127, 113, 124, 105, 75, 57, 48, 36, 0,
		0, 16, -6, -4, 17, 35, 35, 20, -3, -27, -31, -41
	};

	static const double src16[] =
	{
		93, 5, 6, -4, 0, 6, 19, 33, 8, 0, 103, 48, 5, -11, -20, -22, 0, 260,
		-11, 20, 29, 41, 45, 45, 40, 21, 18, 58, 80, 60, 42, 14, -5, -18, 289,
		17, 29, 35, 35, 35, 33, 35, 25, 34, 42, 67, 44, 39, 30, 14, -5, 431,
		34, 47, 54, 59, 57, 49, 56, 27, 88, 96, 98, 73, 34, 11, 0, 5, 821,
		64, 85, 99, 103, 101, 103, 105, 116, 138, 126, 133, 113, 88, 70, 61,
		48, 0, 2, 18, 0, -5, -6, -11, 10, -5, 18, 36, 35, 19, -4, -28, -31,
		-41
	};

	static const double src64[] =
	{
		124, 0, 0, 0, 0, 0, 0, 0, 0, 121, 134, 132, 108, 88, 101, 117, 88,
		33, 37, 22, 40, 30, 27, 29, 4, -2, -7, -15, -11, -3, -17, -8, -16,
		-23, -22, -23, -12, -17, -25, -26, -39, -23, -22, -26, -41, -33, -32,
		-11, -36, -35, -20, -36, -62, -56, -23, -9, -47, 0, 0, 0, 0, 0, 0,
		0, 0, 296, -79, -10, -4, -6, 6, -5, -3, -36, -33, -1, 33, 51, 16, 46,
		-5, -13, -9, 26, 45, 61, 55, 53, 54, -1, -1, 12, 52, 58, 48, 59, 33,
		29, -13, 11, 28, 29, 28, 37, 30, 0, -26, 0, 7, 19, 19, 16, 15, -17,
		-50, -27, -13, -4, -3, -2, -36, -22, -35, -35, -26, -34, -21, -32,
		-27, -49, 331, 0, -4, 7, -6, 0, -22, -13, -8, -31, 12, 17, 4, 14, 6,
		11, -10, -10, 26, 18, 33, 25, 40, 30, 24, -5, 14, 23, 40, 39, 25, 20,
		16, 0, 6, 30, 26, 30, 20, 11, -15, -3, 16, 6, 14, 0, 18, 14, -2, -25,
		-3, 4, -11, 4, 0, 10, -10, -50, -27, -28, -11, -28, -28, -24, -23,
		496, 31, 30, 38, 32, 34, 33, 31, 33, 42, 43, 44, 42, 43, 37, 30, 34,
		32, 35, 37, 39, 35, 42, 28, 20, 10, 12, 30, 24, 18, 23, 11, 2, -17,
		-4, 6, 1, 6, 0, 1, -18, -31, -17, -17, -18, -8, 0, -9, -33, -38, -13,
		-13, -16, -21, -16, -25, -55, -33, -21, -13, -1, -2, -3, -29, -40,
		990, 0, 26, 40, 49, 69, 57, 43, 32, -25, -10, 22, 51, 62, 67, 52, 37,
		-21, 8, 21, 47, 69, 87, 76, 53, -7, -7, 10, 27, 35, 37, 18, 32, -5,
		-10, 9, 22, 23, 17, 15, 0, -24, -1, 0, 4, 9, 8, 7, 0, -35, -11, 1,
		0, 8, 12, -10, -17, -23, -36, -19, -2, -10, -49, -46, -25, 0, 0, -58,
		0, 4, -11, -2, 5, -13, -27, -3, 3, 10, 9, 12, 6, -4, 13, 7, 12, 17,
		18, 15, 11, 0, 14, 13, 19, 21, 18, 12, 1, 4, 0, 10, 13, 11, 15, 15,
		12, 0, -14, 0, -1, 0, 0, 4, 1, -10, -3, -5, -8, -19, -22, -4, -1, -13,
		-25, -2, 3, -39, -8, -31, 11, -24
	};

	switch (PSQ_LENGTH)
	{
	case 5:
		memcpy(&(x[0]), src5, NUMBER_OF_FEATURES * sizeof(double));
		break;
	case 12:
		memcpy(&(x[0]), src12, NUMBER_OF_FEATURES * sizeof(double));
		break;
	case 16:
		memcpy(&(x[0]), src16, NUMBER_OF_FEATURES * sizeof(double));
		break;
	case 64:
		memcpy(&(x[0]), src64, NUMBER_OF_FEATURES * sizeof(double));
		break;
	default:
		break;
	}
}
////////////////////////////////////////////////////////////////////////////////

bool ReadWeights(vector<double>& x, const string& file)
{
	x.clear();
	x.resize(NUMBER_OF_FEATURES);

	ifstream ifs(file.c_str());
	if (ifs.good())
	{
		string s;
		while (getline(ifs, s))
		{
			istringstream iss(s);
			string name;
			double value;

			if (!(iss >> name))
				continue;

			for (size_t i = 0; i < g_features.size(); ++i)
			{
				if (g_features[i].name == name)
				{
					size_t index = g_features[i].index;
					size_t offset = 0;
					while (iss >> value)
					{
						x[index + offset] = value;
						++offset;
						if (offset >= g_features[i].len)
							break;
					}
					break;
				}
			}
		}
		return true;
	}
	return false;
}
////////////////////////////////////////////////////////////////////////////////

void WriteWeights(const vector<double>& x, const string& file)
{
	ofstream ofs(file.c_str());
	if (ofs.good())
	{
		for (size_t i = 0; i < g_features.size(); ++i)
		{
			string name = g_features[i].name;
			size_t index = g_features[i].index;
			size_t len = g_features[i].len;

			ofs << setw(maxTagLength + 2) << left << name;
			for (size_t j = 0; j < len; ++j)
				ofs << " " << (int)x[index + j];
			ofs << endl;
		}
	}
}
////////////////////////////////////////////////////////////////////////////////
